package client.renderer;

import java.util.Arrays;

import common.biome.Biome;
import common.init.Blocks;
import common.tileentity.TileEntity;
import common.util.BlockPos;
import common.util.Facing;
import common.util.Vec3i;
import common.world.Chunk;
import common.world.ChunkCache;
import common.world.IWorldAccess;
import common.world.LightType;
import common.world.State;
import common.world.World;

public class RegionRenderCache extends ChunkCache implements IWorldAccess
{
    private static final State DEFAULT_STATE = Blocks.air.getState();
    
    private final World worldObj;
    private final BlockPos position;
    private final boolean empty;
    private int[] combinedLights;
    private State[] blockStates;

    public RegionRenderCache(World worldIn, BlockPos posFromIn, BlockPos posToIn, int subIn)
    {
        super(worldIn, posFromIn, posToIn, subIn);
        this.worldObj = worldIn;
        boolean empty = true;
        for (int i1 = posFromIn.getX() >> 4; i1 <= posToIn.getX() >> 4; ++i1)
        {
            for (int j1 = posFromIn.getZ() >> 4; j1 <= posToIn.getZ() >> 4; ++j1)
            {
                Chunk chunk = this.chunkArray[i1 - this.chunkX][j1 - this.chunkZ];

                if (chunk != null && !chunk.isEmpty(posFromIn.getY(), posToIn.getY()))
                {
                    empty = false;
                }
            }
        }
        this.empty = empty;
        this.position = posFromIn.subtract(new Vec3i(subIn, subIn, subIn));
        int i = 16000;
        this.combinedLights = new int[16000];
        Arrays.fill((int[])this.combinedLights, (int) - 1);
        this.blockStates = new State[16000];
    }

    public TileEntity getTileEntity(BlockPos pos)
    {
        int i = (pos.getX() >> 4) - this.chunkX;
        int j = (pos.getZ() >> 4) - this.chunkZ;
        return this.chunkArray[i][j].getTileEntity(pos, TileEntity.EnumCreateEntityType.QUEUED);
    }

    public int getCombinedLight(BlockPos pos, int lightValue)
    {
        int i = this.getPositionIndex(pos);
        int j = this.combinedLights[i];

        if (j == -1)
        {
            j = this.getLight(pos, lightValue);
            this.combinedLights[i] = j;
        }

        return j;
    }

    public State getState(BlockPos pos)
    {
        int i = this.getPositionIndex(pos);
        State iblockstate = this.blockStates[i];

        if (iblockstate == null)
        {
            iblockstate = this.getBlockStateRaw(pos);
            this.blockStates[i] = iblockstate;
        }

        return iblockstate;
    }

    private State getBlockStateRaw(BlockPos pos)
    {
        if (pos.getY() >= -World.MAX_SIZE_Y && pos.getY() < World.MAX_SIZE_Y)
        {
            int i = (pos.getX() >> 4) - this.chunkX;
            int j = (pos.getZ() >> 4) - this.chunkZ;
            return this.chunkArray[i][j].getState(pos);
        }
        else
        {
            return DEFAULT_STATE;
        }
    }

    private int getPositionIndex(BlockPos p_175630_1_)
    {
        int i = p_175630_1_.getX() - this.position.getX();
        int j = p_175630_1_.getY() - this.position.getY();
        int k = p_175630_1_.getZ() - this.position.getZ();
        return i * 800 + k * 40 + j;
    }
    
    public boolean isEmpty()
    {
        return this.empty;
    }

    public int getLight(BlockPos pos, int lightValue)
    {
        int i = this.getLightForExt(LightType.SKY, pos);
        int j = this.getLightForExt(LightType.BLOCK, pos);

        if (j < lightValue)
        {
            j = lightValue;
        }

        return i << 20 | j << 4;
    }

    public Biome getBiomeGenForCoords(BlockPos pos)
    {
        return this.worldObj.getBiomeGenForCoords(pos);
    }

    private int getLightForExt(LightType p_175629_1_, BlockPos pos)
    {
        if (p_175629_1_ == LightType.SKY && this.worldObj.dimension.hasNoLight())
        {
            return 0;
        }
        else if (pos.getY() >= -World.MAX_SIZE_Y && pos.getY() < World.MAX_SIZE_Y)
        {
            if (this.getState(pos).getBlock().getSumBrightness())
            {
                int l = 0;

                for (Facing enumfacing : Facing.values())
                {
                    int k = this.getLightFor(p_175629_1_, pos.offset(enumfacing));

                    if (k > l)
                    {
                        l = k;
                    }

                    if (l >= 15)
                    {
                        return l;
                    }
                }

                return l;
            }
            else
            {
                int i = (pos.getX() >> 4) - this.chunkX;
                int j = (pos.getZ() >> 4) - this.chunkZ;
                return this.chunkArray[i][j].getLight(p_175629_1_, pos);
            }
        }
        else
        {
            return p_175629_1_.defValue;
        }
    }

//    /**
//     * Checks to see if an air block exists at the provided location. Note that this only checks to see if the blocks
//     * material is set to air, meaning it is possible for non-vanilla blocks to still pass this check.
//     */
//    public boolean isAirBlock(BlockPos pos)
//    {
//        return this.getBlockState(pos).getBlock().getMaterial() == Material.air;
//    }

    public int getLightFor(LightType p_175628_1_, BlockPos pos)
    {
        if (pos.getY() >= -World.MAX_SIZE_Y && pos.getY() < World.MAX_SIZE_Y)
        {
            int i = (pos.getX() >> 4) - this.chunkX;
            int j = (pos.getZ() >> 4) - this.chunkZ;
            return this.chunkArray[i][j].getLight(p_175628_1_, pos);
        }
        else
        {
            return p_175628_1_.defValue;
        }
    }

//    public int getStrongPower(BlockPos pos, EnumFacing direction)
//    {
//        IBlockState iblockstate = this.getBlockState(pos);
//        return iblockstate.getBlock().getStrongPower(this, pos, iblockstate, direction);
//    }
}
